part of web_api;
/// The core element of the library
/// It analyzes each [HttpRequest] and calls the methods of the controllers
class RequestMapper{

  List<_RequestMapping> mappings = new List<_RequestMapping>();
  Serializer serializer;
  _ParameterResolver parameterResolver = new _ParameterResolver(UTF8, new Serializer.Json());
  ResponseBuilder responseBuilder = new ResponseBuilder(new Serializer.Json(),new LinkAppander());
  Func1<Type,Controller> resolveController;

  Symbol _getBodyParameter(List<ParameterMirror> parameters){
    for(ParameterMirror parameter in parameters){
      for(InstanceMirror metadata in parameter.metadata){
        if(metadata.reflectee is Body){
          return parameter.simpleName;
        }
      }
    }
    return null;
  }

  List<Symbol> _getQueryParameters(List<ParameterMirror> parameters){
    List<Symbol> result = new List();
    for(ParameterMirror parameter in parameters){
      for(InstanceMirror metadata in parameter.metadata){
        if(metadata.reflectee is Query){
          result.add(parameter.simpleName);
        }
      }
    }
    return result;
  }

  /// Registers a controller
  /// You have to register each controller you want to use to handle a [HttpRequest]
  /// @param controller the Type of the controller you want to register
  void addController(Type controller){
    ClassMirror classMirror = reflectType(controller);
    for(DeclarationMirror declaration in classMirror.declarations.values){
      if(declaration is MethodMirror){
        for(InstanceMirror metadata in declaration.metadata){
          if(metadata.reflectee is APIMethod){
            APIMethod apiMehtod = metadata.reflectee;
            Symbol bodyParameter = _getBodyParameter(declaration.parameters);
            List<Symbol> queryParameters = _getQueryParameters(declaration.parameters);
            _RequestMapping mapping = new _RequestMapping(apiMehtod.url,apiMehtod.method,controller,bodyParameter,declaration.simpleName,apiMehtod.successStatus,queryParameters);
            mappings.add(mapping);
          }
        }
      }
    }
  }

  Future _callController(HttpRequest request, _RequestMapping mapping, Controller controllerInstance) async{
    List<Object> paramList = await parameterResolver.resolve(request,mapping);
    InstanceMirror controllerInstanceMirror = reflect(controllerInstance);
    Object result = await new Future.value(controllerInstanceMirror.invoke(mapping.controllerMethod,paramList).reflectee);
    return result;
  }

  Future _handleRequest(HttpRequest request) async{
    List<_RequestMapping> foundMappings = mappings.where((m){
      String pattern ='';
      Uri mappingUri = new  Uri(path: '');
      for(String segment in m.url.split("\/")){
        if(segment.length==0){
          continue;
        }
        if(segment.startsWith('{') && segment.endsWith('}')){
          pattern+=r"\/\w+";
        }
        else{
          pattern+=r'\/'+segment;
        }
      }
      pattern+=r"$";
      if(pattern==r'$'){
          pattern=r'/$';
      }
      RegExp regex = new RegExp(pattern);
      bool match = regex.hasMatch(request.uri.path) && request.method==m.method;
      return match;
    }).toList();
    if(foundMappings.length==0 && request.method!="OPTIONS"){
      await responseBuilder.unknownRoute(request);
      return;
    }
    else if(foundMappings.length==0){
      await responseBuilder.allowOption(request);
      return;
    }
    _RequestMapping mapping = foundMappings[0];
    try{
        Controller controllerInstance = resolveController(mapping.controller);
        controllerInstance.request=request;
        Object controllerResult = await _callController(request,mapping,controllerInstance);
        await responseBuilder.success(request,mapping,controllerResult,controllerInstance);
    }catch(e){
      if(e is ControllerException){
        await responseBuilder.controllerException(request,e);
      }
      else{
        await responseBuilder.exception(request,e);
      }
    }

  }

  /// Starts to handle all requests
  /// @param server The [HttpServer] which requests should be handled
  Future handle(HttpServer server)async{
    await for(HttpRequest request in server){
      _handleRequest(request);
    }
  }

  /// @param resolveController A method, which returns an instance of the provided type
  /// * The provided type always extends [Controller]
  /// * This method is required to create a new controller for each request
  RequestMapper(Func1<Type,Controller> resolveController){
    serializer=new Serializer.Json();
    this.resolveController=resolveController;
  }

}
